#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import mock.mock
import pytest
import octobot_trading.api as api
import octobot_trading.exchanges as exchanges
import octobot_commons.authentication as commons_authentication

from tests.exchanges import exchange_manager
from tests import event_loop


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_build_trading_modes_if_required(exchange_manager):
    builder = exchanges.ExchangeBuilder({}, "binanceus")
    builder.exchange_manager = exchange_manager
    tentacles_setup_config = exchange_manager.tentacles_setup_config

    # no set trader: no trading mode creation attempt
    assert builder.exchange_manager.trader is None
    await builder._build_trading_modes_if_required(None, tentacles_setup_config)
    assert builder.exchange_manager.trader is None

    # with trader simulator: will attempt to create a trading mode and fail (because of the None arg)
    builder.is_simulated()
    with mock.patch.object(
        commons_authentication.Authenticator, "has_open_source_package", mock.Mock(return_value=False)
    ) as has_open_source_package_mock:
        # get_is_using_trading_mode_on_exchange returns True
        trading_mode_class = mock.Mock(
            get_is_using_trading_mode_on_exchange=mock.Mock(return_value=True),
            get_supported_exchange_types=mock.Mock(return_value=[]),
            return_value=mock.Mock(
                initialize=mock.AsyncMock(),
            ),
        )
        with pytest.raises(AttributeError):
            await builder._build_trading_modes_if_required(None, tentacles_setup_config)
        has_open_source_package_mock.assert_not_called()
        await builder._build_trading_modes_if_required(trading_mode_class, tentacles_setup_config)
        trading_mode_class.get_is_using_trading_mode_on_exchange.assert_called_once_with(builder.exchange_name, tentacles_setup_config)
        trading_mode_class.get_supported_exchange_types.assert_called_once()
        trading_mode_class.return_value.initialize.assert_awaited_once()
        has_open_source_package_mock.assert_called_once()
        trading_mode_class.reset_mock()
        trading_mode_class.get_is_using_trading_mode_on_exchange.reset_mock()
        trading_mode_class.get_supported_exchange_types.reset_mock()
        trading_mode_class.return_value.initialize.reset_mock()

        # get_is_using_trading_mode_on_exchange returns False
        trading_mode_class.get_is_using_trading_mode_on_exchange.return_value = False
        await builder._build_trading_modes_if_required(trading_mode_class, tentacles_setup_config)
        trading_mode_class.get_is_using_trading_mode_on_exchange.assert_called_once_with(builder.exchange_name, tentacles_setup_config)
        trading_mode_class.get_supported_exchange_types.assert_not_called()
        trading_mode_class.return_value.initialize.assert_not_called()
        has_open_source_package_mock.assert_called_once()

    # raised by default has_open_source_package (which should be overriden)
    trading_mode_class.get_is_using_trading_mode_on_exchange.return_value = True
    with pytest.raises(NotImplementedError):
        await builder._build_trading_modes_if_required(trading_mode_class, tentacles_setup_config)


async def test_build_collector_exchange(exchange_manager):
    new_exchange_manager = await api.create_exchange_builder(exchange_manager.config, exchange_manager.exchange_name) \
        .is_simulated() \
        .is_rest_only() \
        .is_exchange_only() \
        .is_without_auth() \
        .is_ignoring_config() \
        .disable_trading_mode() \
        .use_tentacles_setup_config(exchange_manager.tentacles_setup_config) \
        .build()
    assert new_exchange_manager is not exchange_manager
    assert new_exchange_manager.config is exchange_manager.config
    assert new_exchange_manager.exchange_name == exchange_manager.exchange_name
    assert new_exchange_manager.tentacles_setup_config is exchange_manager.tentacles_setup_config
    await new_exchange_manager.stop()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import mock
import pytest
import octobot_trading.personal_data as personal_data
import octobot_trading.enums as enums


@pytest.fixture
def take_profit_first_strategy():
    return personal_data.TakeProfitFirstActiveOrderSwapStrategy(
        123, enums.ActiveOrderSwapTriggerPriceConfiguration.FILLING_PRICE.value
    )


def test_is_priority_order(take_profit_first_strategy):
    assert take_profit_first_strategy.swap_timeout == 123
    assert take_profit_first_strategy.is_priority_order(mock.Mock(order_type=enums.TraderOrderType.STOP_LOSS)) is False
    assert take_profit_first_strategy.is_priority_order(mock.Mock(order_type=enums.TraderOrderType.SELL_LIMIT)) is True
    assert take_profit_first_strategy.is_priority_order(mock.Mock(order_type=enums.TraderOrderType.BUY_LIMIT)) is True



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import contextlib
import dotenv
import os
import mock

import octobot_commons.constants as commons_constants
import octobot_commons.asyncio_tools as asyncio_tools
import octobot_commons.tests.test_config as test_config
import octobot_trading.api as api
import octobot_trading.exchanges as exchanges
import octobot_trading.enums as enums
import octobot_trading.errors as errors
import octobot_trading.exchanges.util.exchange_util


LOADED_EXCHANGE_CREDS_ENV_VARIABLES = False


@contextlib.asynccontextmanager
async def get_exchange_manager(exchange_name, config=None, authenticated=False, market_filter=None, uses_tentacle=False):
    config = {**test_config.load_test_config(), **config} if config else test_config.load_test_config()
    if exchange_name not in config[commons_constants.CONFIG_EXCHANGES]:
        config[commons_constants.CONFIG_EXCHANGES][exchange_name] = {}
    if authenticated:
        config[commons_constants.CONFIG_EXCHANGES][exchange_name].update(_get_exchange_auth_details(
            exchange_name
        ))
    exchange_manager_instance = exchanges.ExchangeManager(config, exchange_name)
    exchange_manager_instance.market_filter = market_filter
    if config[commons_constants.CONFIG_EXCHANGES][exchange_name]. \
       get(commons_constants.CONFIG_EXCHANGE_TYPE, enums.ExchangeTypes.SPOT.value) == enums.ExchangeTypes.FUTURE.value:
        exchange_manager_instance.is_future = True
    if not uses_tentacle:
        # don't use exchange specific tentacles, even if they are available
        with mock.patch.object(
            octobot_trading.exchanges.util.exchange_util, 
            "search_exchange_class_from_exchange_name", 
            mock.Mock(return_value=exchanges.DefaultRestExchange)
        ) as search_exchange_class_from_exchange_name_mock:
            await exchange_manager_instance.initialize(exchange_config_by_exchange=None)
            assert search_exchange_class_from_exchange_name_mock.call_count > 0
    else:
        # allow the use of exchange specific tentacles
        await exchange_manager_instance.initialize(exchange_config_by_exchange=None)
    try:
        yield exchange_manager_instance
    except errors.UnreachableExchange as err:
        raise errors.UnreachableExchange(f"{exchange_name} can't be reached, it is either offline or you are not connected "
                                         "to the internet (or a proxy is preventing connecting to this exchange).") \
            from err
    finally:
        await exchange_manager_instance.stop()
        api.cancel_ccxt_throttle_task()
        # let updaters gracefully shutdown
        await asyncio_tools.wait_asyncio_next_cycle()


def _load_exchange_creds_env_variables_if_necessary():
    global LOADED_EXCHANGE_CREDS_ENV_VARIABLES
    if not LOADED_EXCHANGE_CREDS_ENV_VARIABLES:
        # load environment variables from .env file if exists
        dotenv_path = os.getenv("EXCHANGE_TESTS_DOTENV_PATH", os.path.dirname(os.path.abspath(__file__)))
        dotenv.load_dotenv(os.path.join(dotenv_path, ".env"), verbose=False)
        LOADED_EXCHANGE_CREDS_ENV_VARIABLES = True


def _get_exchange_auth_details(exchange_name):
    _load_exchange_creds_env_variables_if_necessary()
    return {
        commons_constants.CONFIG_EXCHANGE_KEY:
            _get_exchange_credential_from_env(exchange_name, commons_constants.CONFIG_EXCHANGE_KEY),
        commons_constants.CONFIG_EXCHANGE_SECRET:
            _get_exchange_credential_from_env(exchange_name, commons_constants.CONFIG_EXCHANGE_SECRET),
        commons_constants.CONFIG_EXCHANGE_PASSWORD:
            _get_exchange_credential_from_env(exchange_name, commons_constants.CONFIG_EXCHANGE_PASSWORD),
    }


def _get_exchange_credential_from_env(exchange_name, cred_suffix):
    # for bybit api key: get BYBIT_KEY (as encrypted value)
    # for bybit api password: get BYBIT_PASSWORD (as encrypted value)
    # for bybit api secret: get BYBIT_SECRET (as encrypted value)
    return os.getenv(f"{exchange_name}_{cred_suffix.split('-')[-1]}".upper(), None)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.logging as logging
import octobot_trading.enums as enums
import octobot_trading.personal_data.orders.order_factory as order_factory
import octobot_trading.personal_data.orders.groups.group_util as group_util


LOGGER_NAME = "orders_storage_operations"


async def apply_order_storage_details_if_any(order, exchange_manager, pending_groups):
    # only real orders can be updated by stored orders
    if not exchange_manager.storage_manager.orders_storage \
            or not exchange_manager.storage_manager.orders_storage.should_store_data():
        return
    order_details = await exchange_manager.storage_manager.orders_storage.get_startup_order_details(
        order.exchange_order_id
    )
    if order_details:
        logging.get_logger(LOGGER_NAME).debug(f"Updating fetched order {order} using stored order details")
        order.update_from_storage_order_details(order_details)
        await create_orders_storage_related_elements(order, order_details, exchange_manager, pending_groups)


async def create_orders_storage_related_elements(order, order_storage_details, exchange_manager, pending_groups):
    group = group_util.get_or_create_order_group_from_storage_order_details(order_storage_details, exchange_manager)
    if group:
        order.add_to_order_group(group)
        logging.get_logger(LOGGER_NAME).debug(f"Adding {order} to restored group {group}")
        pending_groups[group.name] = group
    await order_factory.restore_chained_orders_from_storage_order_details(
        order, order_storage_details, exchange_manager, pending_groups
    )


async def create_order_from_storage_data(order_desc: dict, exchange_manager, pending_groups):
    created_order = await order_factory.create_order_from_order_storage_details(
        order_desc, exchange_manager, pending_groups
    )
    await created_order.initialize()
    return created_order


async def _create_storage_virtual_orders_from_group(pending_group_id, exchange_manager, pending_groups):
    try:
        to_create_orders = exchange_manager.storage_manager.orders_storage \
            .get_startup_virtual_orders_details_from_group(pending_group_id)
        for order_desc in to_create_orders:
            await create_order_from_storage_data(order_desc, exchange_manager, pending_groups)
    except Exception as err:
        logging.get_logger(LOGGER_NAME).exception(
            err, True, f"Error when creating {pending_group_id} group virtual orders with stored data: {err}"
        )


async def create_required_virtual_orders(pending_groups, exchange_manager):
    virtual_orders = exchange_manager.storage_manager.orders_storage.get_all_virtual_startup_orders()
    # virtual order that are not within a group will only be restored here
    virtual_non_grouped_order_details = [
        order
        for order in virtual_orders
        if order.get(enums.StoredOrdersAttr.GROUP.value, {}) == {}
    ]
    logger = logging.get_logger(LOGGER_NAME)
    for order_details in virtual_non_grouped_order_details:
        try:
            order = await create_order_from_storage_data(order_details, exchange_manager, pending_groups)
            logger.debug(f"Restored order from storage: {order}")
        except Exception as err:
            logger.exception(err, True, f"Error when restoring virtual order order using stored data: {err}")


async def create_missing_virtual_orders_from_storage_order_groups(pending_groups, exchange_manager):
    # create order groups' associated self-managed and inactive orders if any
    to_complete_groups = list(pending_groups.keys())
    completed_groups = set()
    max_allowed_nested_chained_orders_groups = 100
    # Loop as created virtual orders might carry chained orders
    # themselves linked to a group with self-managed orders.
    # This would be seen after each iteration only
    # However only loop a max amount of time as a huge looping amount would indicate an error.
    for _ in range(max_allowed_nested_chained_orders_groups):
        for pending_group_id in to_complete_groups:
            await _create_storage_virtual_orders_from_group(pending_group_id, exchange_manager, pending_groups)
        # do not process the same group twice
        completed_groups = completed_groups.union(set(to_complete_groups))
        to_complete_groups = [
            group_id
            for group_id in pending_groups.keys()
            if group_id not in completed_groups
        ]
        if not to_complete_groups:
            return
    # if we arrived here, it means that after 100 iterations, still not every order group is complete,
    # there is an issue.
    logging.get_logger(LOGGER_NAME).error(
        f"Error when completing order groups: {len(to_complete_groups)} remaining order "
        f"groups to complete after maximum iterations."
    )



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import mock
from mock import Mock, AsyncMock
import pytest

import octobot_trading.api as api
import octobot_trading.enums as enums
import octobot_trading.constants as constants
import octobot_trading.errors
import octobot_trading.personal_data as personal_data
import octobot_trading.personal_data.orders.order_util as order_util
import octobot_trading.personal_data.orders.order_factory as order_factory

from tests import event_loop
from tests.exchanges import future_simulated_exchange_manager, simulated_exchange_manager, set_future_exchange_fees
from tests.exchanges.traders import future_trader_simulator_with_default_linear, \
    future_trader_simulator_with_default_inverse, DEFAULT_FUTURE_SYMBOL, DEFAULT_FUTURE_FUNDING_RATE, trader_simulator


def test_get_min_max_amounts():
    # normal values
    symbol_market = {
        enums.ExchangeConstantsMarketStatusColumns.LIMITS.value: {
            enums.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT.value: {
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT_MIN.value: 0.5,
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT_MAX.value: 100,
            },
            enums.ExchangeConstantsMarketStatusColumns.LIMITS_COST.value: {
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_COST_MIN.value: None,
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_COST_MAX.value: None
            },
            enums.ExchangeConstantsMarketStatusColumns.LIMITS_PRICE.value: {
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_PRICE_MIN.value: 0.5,
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_PRICE_MAX.value: 50
            },
        }
    }
    min_quantity, max_quantity, min_cost, max_cost, min_price, max_price = personal_data.get_min_max_amounts(
        symbol_market)
    assert min_quantity == 0.5
    assert max_quantity == 100
    assert min_cost is None
    assert max_cost is None
    assert min_price == 0.5
    assert max_price == 50

    # missing all values
    min_quantity, max_quantity, min_cost, max_cost, min_price, max_price = personal_data.get_min_max_amounts({})
    assert min_quantity is None
    assert max_quantity is None
    assert min_cost is None
    assert max_cost is None
    assert min_price is None
    assert max_price is None

    # missing all values: asign default
    min_quantity, max_quantity, min_cost, max_cost, min_price, max_price = personal_data.get_min_max_amounts({}, "xyz")
    assert min_quantity == "xyz"
    assert max_quantity == "xyz"
    assert min_cost == "xyz"
    assert max_cost == "xyz"
    assert min_price == "xyz"
    assert max_price == "xyz"

    # missing values: assign default

    symbol_market = {
        enums.ExchangeConstantsMarketStatusColumns.LIMITS.value: {
            enums.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT.value: {
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT_MIN.value: 0.5,
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT_MAX.value: 100,
            },
            enums.ExchangeConstantsMarketStatusColumns.LIMITS_COST.value: {
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_COST_MIN.value: None,
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_COST_MAX.value: None
            }
        }
    }
    min_quantity, max_quantity, min_cost, max_cost, min_price, max_price = personal_data.get_min_max_amounts(symbol_market, "xyz")
    assert min_quantity == 0.5
    assert max_quantity == 100
    assert min_cost == "xyz"  # None is not a valid value => assign default
    assert max_cost == "xyz"  # None is not a valid value => assign default
    assert min_price == "xyz"
    assert max_price == "xyz"


def test_get_fees_for_currency():
    fee1 = {
        enums.FeePropertyColumns.CURRENCY.value: "BTC",
        enums.FeePropertyColumns.COST.value: 1
    }
    assert personal_data.get_fees_for_currency(fee1, "BTC") == 1
    assert personal_data.get_fees_for_currency(fee1, "BTC1") == 0

    fee2 = {
        enums.FeePropertyColumns.CURRENCY.value: "BTC",
        enums.FeePropertyColumns.COST.value: 0,
        enums.FeePropertyColumns.IS_FROM_EXCHANGE.value: True
    }
    assert personal_data.get_fees_for_currency(fee2, "BTC") == 0
    assert personal_data.get_fees_for_currency(fee2, "BTC1") == 0

    assert personal_data.get_fees_for_currency({}, "BTC") == 0
    assert personal_data.get_fees_for_currency(None, "BTC") == 0


def test_get_max_order_quantity_for_price_long_linear(future_trader_simulator_with_default_linear):
    # values also tested with bybit fees
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear
    set_future_exchange_fees(exchange_manager_inst.exchange.connector, default_contract.pair)

    # no need to initialize the position
    default_contract.set_current_leverage(constants.ONE)
    # at price = 37000 and 9961.7672 USDT in stock, if there were no fees,
    # max quantity would be 9961.7672 / 37000 = 0.269236951351,
    # it is actually less to allow fees
    assert personal_data.get_max_order_quantity_for_price(
        personal_data.LinearPosition(trader_inst, default_contract),
        decimal.Decimal("9961.7672"), decimal.Decimal("37000"), enums.PositionSide.LONG, DEFAULT_FUTURE_SYMBOL
    ) == decimal.Decimal('0.2691292996314987518506111069')

    default_contract.set_current_leverage(decimal.Decimal("2"))
    # at price = 37000 and 9961.7672 USDT in stock, if there were no fees,
    # max quantity would be 9961.7672 / 37000 * 2 = 0.538473902703,
    # it is actually less to allow fees
    assert personal_data.get_max_order_quantity_for_price(
        personal_data.LinearPosition(trader_inst, default_contract),
        decimal.Decimal("9961.7672"), decimal.Decimal("37000"), enums.PositionSide.LONG, DEFAULT_FUTURE_SYMBOL
    ) == decimal.Decimal('0.5378285084925116886762911533')

    default_contract.set_current_leverage(decimal.Decimal("10"))
    # at price = 43500 and 9943.9078 USDT in stock, if there were no fees,
    # max quantity would be 9943.9078 / 43500 * 10 = 2.26311218138,
    # it is actually less to allow fees
    assert personal_data.get_max_order_quantity_for_price(
        personal_data.LinearPosition(trader_inst, default_contract),
        decimal.Decimal("9943.9078"), decimal.Decimal("43500"), enums.PositionSide.LONG, DEFAULT_FUTURE_SYMBOL
    ) == decimal.Decimal('2.268713592786774536511021980')

    # example from bybit docs
    # Trader place a long entry of 1BTC at USD8000 with 50x leverage.
    # Order Cost = 160USDT + 6USDT + 5.88USDT = 171.88 USDT
    default_contract.set_current_leverage(decimal.Decimal("50"))
    assert personal_data.get_max_order_quantity_for_price(
        personal_data.LinearPosition(trader_inst, default_contract),
        decimal.Decimal("171.88"), decimal.Decimal("8000"), enums.PositionSide.LONG, DEFAULT_FUTURE_SYMBOL
    ) == decimal.Decimal('1.033330126971912273951519815')
    # Here the max size is a bit higher than 1 since binance fees are a bit higher

    default_contract.set_current_leverage(constants.ONE_HUNDRED)
    # at price = 43500 and 9943.9078 USDT in stock, if there were no fees,
    # max quantity would be 9943.9078 / 43500 * 100 = 22.6311218138,
    # it is actually less to allow fees (which are huge on 100x)
    assert personal_data.get_max_order_quantity_for_price(
        personal_data.LinearPosition(trader_inst, default_contract),
        decimal.Decimal("9943.9078"), decimal.Decimal("43500"), enums.PositionSide.LONG, DEFAULT_FUTURE_SYMBOL
    ) == decimal.Decimal('21.17409981559794389578089799')


def test_get_max_order_quantity_for_price_short_linear(future_trader_simulator_with_default_linear):
    # values also tested with bybit fees
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear
    set_future_exchange_fees(exchange_manager_inst.exchange.connector, default_contract.pair)

    # no need to initialize the position
    # Differs from long due to the position closing fees at liquidation price which are higher (price is higher)
    # Therefore to open the same position size, more funds are required than for longs
    # (max quantity is lower than for longs)
    default_contract.set_current_leverage(constants.ONE)
    # at price = 37000 and 9961.7672 USDT in stock, if there were no fees,
    # max quantity would be 9961.7672 / 37000 = 0.269236951351,
    # it is actually less to allow fees
    assert personal_data.get_max_order_quantity_for_price(
        personal_data.LinearPosition(trader_inst, default_contract),
        decimal.Decimal("9961.7672"), decimal.Decimal("37000"), enums.PositionSide.SHORT, DEFAULT_FUTURE_SYMBOL
    ) == decimal.Decimal('0.2689142542462558443381455767')

    default_contract.set_current_leverage(decimal.Decimal("2"))
    # at price = 37000 and 9961.7672 USDT in stock, if there were no fees,
    # max quantity would be 9961.7672 / 37000 * 2 = 0.538473902703,
    # it is actually less to allow fees
    assert personal_data.get_max_order_quantity_for_price(
        personal_data.LinearPosition(trader_inst, default_contract),
        decimal.Decimal("9961.7672"), decimal.Decimal("37000"), enums.PositionSide.SHORT, DEFAULT_FUTURE_SYMBOL
    ) == decimal.Decimal('0.5373991044937152721583859308')

    default_contract.set_current_leverage(decimal.Decimal("50"))
    # at price = 44500 and 9943.9078 USDT in stock, if there were no fees,
    # max quantity would be 9943.9078 / 44500 * 50 = 11.1729301124,
    # it is actually less to allow fees
    assert personal_data.get_max_order_quantity_for_price(
        personal_data.LinearPosition(trader_inst, default_contract),
        decimal.Decimal("9943.9078"), decimal.Decimal("44500"), enums.PositionSide.SHORT, DEFAULT_FUTURE_SYMBOL
    ) == decimal.Decimal('10.73907161895381638004397617')

    default_contract.set_current_leverage(constants.ONE_HUNDRED)
    # at price = 37000 and 9961.7672 USDT in stock, if there were no fees,
    # max quantity would be 9961.7672 / 37000 * 100 = 26.9236951351,
    # it is actually less to allow fees (which are huge on 100x)
    assert personal_data.get_max_order_quantity_for_price(
        personal_data.LinearPosition(trader_inst, default_contract),
        decimal.Decimal("9961.7672"), decimal.Decimal("37000"), enums.PositionSide.SHORT, DEFAULT_FUTURE_SYMBOL
    ) == decimal.Decimal('24.92011767413470486406436055')


def test_get_max_order_quantity_for_price_long_inverse(future_trader_simulator_with_default_inverse):
    # values also tested with bybit fees
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_inverse
    set_future_exchange_fees(exchange_manager_inst.exchange.connector, default_contract.pair)

    # no need to initialize the position
    default_contract.set_current_leverage(constants.ONE)
    # at price = 36000 and 1 btc in stock, if there were no fees,
    # max quantity would be 1 * 36000 = 36000,
    # it is actually less to allow fees
    assert personal_data.get_max_order_quantity_for_price(
        personal_data.InversePosition(trader_inst, default_contract),
        constants.ONE, decimal.Decimal("36000"), enums.PositionSide.LONG, DEFAULT_FUTURE_SYMBOL
    ) == decimal.Decimal('35956.85177786656012784658410')

    default_contract.set_current_leverage(decimal.Decimal("2"))
    # at price = 36000 and 1 btc in stock, if there were no fees,
    # max quantity would be 1 * 36000 * 2 = 72000,
    # it is actually less to allow fees
    assert personal_data.get_max_order_quantity_for_price(
        personal_data.InversePosition(trader_inst, default_contract),
        constants.ONE, decimal.Decimal("36000"), enums.PositionSide.LONG, DEFAULT_FUTURE_SYMBOL
    ) == decimal.Decimal('71856.28742514970059880239519')

    default_contract.set_current_leverage(decimal.Decimal("25"))
    # example from bybit docs
    # Trader places a buy limit order of 10,000 BTCUSD contracts at 6,400 USD, using 25x leverage.
    # Order Cost = 0.0625 BTC (Initial margin) + 0.00117188 BTC (fee to open) + 0.00121872 BTC (fee to close) = 0.06489060 BTC
    assert personal_data.get_max_order_quantity_for_price(
        personal_data.InversePosition(trader_inst, default_contract),
        decimal.Decimal("0.06489060"), decimal.Decimal("6400"), enums.PositionSide.LONG, DEFAULT_FUTURE_SYMBOL
    ) == decimal.Decimal('10174.92747941983535868286946')
    # higher than the 10000 from bybit example as binance fees are lower

    default_contract.set_current_leverage(constants.ONE_HUNDRED)
    # at price = 36000 and 1 btc in stock, if there were no fees,
    # max quantity would be 1 * 36000 * 100 = 3600000,
    # it is actually less to allow fees (which are huge on 100x)
    assert personal_data.get_max_order_quantity_for_price(
        personal_data.InversePosition(trader_inst, default_contract),
        constants.ONE, decimal.Decimal("36000"), enums.PositionSide.LONG, DEFAULT_FUTURE_SYMBOL
    ) == decimal.Decimal('3332099.222510181414291003332')


def test_get_max_order_quantity_for_price_short_inverse(future_trader_simulator_with_default_inverse):
    # values also tested with bybit fees
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_inverse
    set_future_exchange_fees(exchange_manager_inst.exchange.connector, default_contract.pair)

    # no need to initialize the position
    # Differs from long due to the position closing fees at liquidation price which are higher (price is higher)
    # Therefore to open the same position size, less funds are required than for longs
    # (max quantity is higher than for longs)
    default_contract.set_current_leverage(constants.ONE)
    # at price = 36000 and 1 btc in stock, if there were no fees,
    # max quantity would be 1 * 36000 = 36000,
    # it is actually less to allow fees
    assert personal_data.get_max_order_quantity_for_price(
        personal_data.InversePosition(trader_inst, default_contract),
        constants.ONE, decimal.Decimal("36000"), enums.PositionSide.SHORT, DEFAULT_FUTURE_SYMBOL
    ) == decimal.Decimal('35985.60575769692123150739704')

    default_contract.set_current_leverage(decimal.Decimal("2"))
    # at price = 36000 and 1 btc in stock, if there were no fees,
    # max quantity would be 1 * 36000 * 2 = 72000,
    # it is actually less to allow fees
    assert personal_data.get_max_order_quantity_for_price(
        personal_data.InversePosition(trader_inst, default_contract),
        constants.ONE, decimal.Decimal("36000"), enums.PositionSide.SHORT, DEFAULT_FUTURE_SYMBOL
    ) == decimal.Decimal('71913.70355573312025569316820')

    default_contract.set_current_leverage(constants.ONE_HUNDRED)
    # at price = 36000 and 1 btc in stock, if there were no fees,
    # max quantity would be 1 * 36000 * 100 = 3600000,
    # it is actually less to allow fees (which are huge on 100x)
    assert personal_data.get_max_order_quantity_for_price(
        personal_data.InversePosition(trader_inst, default_contract),
        constants.ONE, decimal.Decimal("36000"), enums.PositionSide.SHORT, DEFAULT_FUTURE_SYMBOL
    ) == decimal.Decimal('3334568.358651352352723230826')


@pytest.mark.asyncio
async def test_get_futures_max_order_size(future_trader_simulator_with_default_linear):
    # values also tested with bybit fees
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear
    set_future_exchange_fees(exchange_manager_inst.exchange.connector, default_contract.pair)

    symbol = default_contract.pair
    api.force_set_mark_price(exchange_manager_inst, symbol, 37000)
    compare_accuracy = decimal.Decimal("1.000000")

    default_contract.set_current_leverage(constants.ONE)
    l1_current_symbol_holding, _, l1_market_quantity, _, _ = \
        await personal_data.get_pre_order_data(exchange_manager_inst, symbol=symbol)
    assert personal_data.get_futures_max_order_size(
        exchange_manager_inst, symbol, enums.TradeOrderSide.BUY,
        decimal.Decimal("37000"), False, l1_current_symbol_holding, l1_market_quantity
    ) == (decimal.Decimal('0.02701622053881150242605660439'), True)

    # take leverage into account
    default_contract.set_current_leverage(decimal.Decimal(5))
    l2_current_symbol_holding, _, l2_market_quantity, _, _ = \
        await personal_data.get_pre_order_data(exchange_manager_inst, symbol=symbol)
    assert l2_current_symbol_holding.quantize(compare_accuracy) == \
           (l1_current_symbol_holding * default_contract.current_leverage).quantize(compare_accuracy)
    assert l2_market_quantity.quantize(compare_accuracy) == \
           (l1_market_quantity * default_contract.current_leverage).quantize(compare_accuracy)
    full_buy_size, increasing = personal_data.get_futures_max_order_size(
        exchange_manager_inst, symbol, enums.TradeOrderSide.BUY,
        decimal.Decimal("37000"), False, l2_current_symbol_holding, l2_market_quantity
    )
    assert (full_buy_size, increasing) == (decimal.Decimal('0.1346503937177512307045985802'), True)

    # reduce only (position is empty)
    assert personal_data.get_futures_max_order_size(
        exchange_manager_inst, symbol, enums.TradeOrderSide.BUY,
        decimal.Decimal("37000"), True, l2_current_symbol_holding, l2_market_quantity
    ) == (constants.ZERO, False)

    # with short position
    full_sell_size, increasing = personal_data.get_futures_max_order_size(
        exchange_manager_inst, symbol, enums.TradeOrderSide.SELL,
        decimal.Decimal("37000"), False, l2_current_symbol_holding, l2_market_quantity
    )
    assert (full_sell_size, increasing) == (decimal.Decimal('0.1345431452958334678764786291'), True)
    # reduce only (position is empty)
    assert personal_data.get_futures_max_order_size(
        exchange_manager_inst, symbol, enums.TradeOrderSide.SELL,
        decimal.Decimal("37000"), True, l2_current_symbol_holding, l2_market_quantity
    ) == (constants.ZERO, False)

    # LONG POSITION
    # reduce only (position is NOT empty)
    position = exchange_manager_inst.exchange_personal_data.positions_manager.get_symbol_positions(symbol)[0]
    await position.update(mark_price=decimal.Decimal(37000), update_margin=decimal.Decimal(4))
    # position is now long
    position_size = position.size
    assert position_size == decimal.Decimal("0.0005405405405405405405405405405")
    assert personal_data.get_futures_max_order_size(
        exchange_manager_inst, symbol, enums.TradeOrderSide.SELL,
        decimal.Decimal("37000"), True, l2_current_symbol_holding, l2_market_quantity
    ) == (position_size, False)

    # not reduce only and long position: can reverse position
    position_size = position.size
    assert position_size == decimal.Decimal("0.0005405405405405405405405405405")    # still the same size
    max_size, increasing = personal_data.get_futures_max_order_size(
        exchange_manager_inst, symbol, enums.TradeOrderSide.SELL,
        decimal.Decimal("37000"), False, l2_current_symbol_holding, l2_market_quantity
    )
    assert increasing is True
    assert max_size > position_size
    # higher than when just opening a position with full free funds
    assert max_size > full_sell_size
    assert max_size == decimal.Decimal("0.1352986545732659722297027996")

    # SHORT POSITION
    # reduce only (position is NOT empty)
    await position.update(mark_price=decimal.Decimal(37000), update_margin=decimal.Decimal(-500))
    # position is now short
    position_size = position.size
    assert position_size == decimal.Decimal("-0.06702702702702702702702702703")
    assert personal_data.get_futures_max_order_size(
        exchange_manager_inst, symbol, enums.TradeOrderSide.BUY,
        decimal.Decimal("37000"), True, l2_current_symbol_holding, l2_market_quantity
    ) == (-position_size, False)

    # not reduce only and short position: can reverse position
    position_size = position.size
    assert position_size == decimal.Decimal("-0.06702702702702702702702702703")    # still the same size
    max_size, increasing = personal_data.get_futures_max_order_size(
        exchange_manager_inst, symbol, enums.TradeOrderSide.BUY,
        decimal.Decimal("37000"), False, l2_current_symbol_holding, l2_market_quantity
    )
    assert increasing is True
    assert max_size > abs(position_size)
    # higher than when just opening a position with full free funds
    assert max_size > full_sell_size
    assert max_size == decimal.Decimal("0.2149168523362071749168523362")


@pytest.mark.asyncio
async def test_create_as_chained_order_regular_order(trader_simulator):
    config, exchange_manager_inst, trader_inst = trader_simulator

    base_order = personal_data.BuyLimitOrder(trader_inst)
    base_order.update(order_type=enums.TraderOrderType.BUY_LIMIT,
                      symbol="BTC/USDT",
                      current_price=decimal.Decimal("70"),
                      quantity=decimal.Decimal("10"),
                      price=decimal.Decimal("70"))
    base_order.is_waiting_for_chained_trigger = True
    base_order.creation_time = 123
    origin_time = base_order.creation_time
    assert base_order.to_dict()[enums.ExchangeConstantsOrderColumns.TIMESTAMP.value] \
           == origin_time
    # base_order.state is None since no state has been associated to it (not initiliazed)
    assert base_order.state is None
    assert base_order.is_initialized is False

    await personal_data.create_as_chained_order(base_order)
    assert base_order.creation_time > origin_time   # creation_time got update using current time
    assert base_order.is_waiting_for_chained_trigger is False
    assert base_order.is_created() is True
    assert base_order.is_initialized is True
    assert isinstance(base_order.state, personal_data.OpenOrderState)

    # use chained order creation time
    assert base_order.to_dict()[enums.ExchangeConstantsOrderColumns.TIMESTAMP.value] \
           == base_order.creation_time


@pytest.mark.asyncio
async def test_create_as_chained_order_inactive_order(trader_simulator):
    config, exchange_manager_inst, trader_inst = trader_simulator
    trader_inst.enable_inactive_orders = True

    base_order = personal_data.BuyLimitOrder(trader_inst)
    base_order.update(order_type=enums.TraderOrderType.BUY_LIMIT,
                      symbol="BTC/USDT",
                      current_price=decimal.Decimal("70"),
                      quantity=decimal.Decimal("10"),
                      price=decimal.Decimal("70"))
    trigger = order_util.create_order_price_trigger(base_order, decimal.Decimal("90"), True)
    with mock.patch.object(trigger, "create_watcher", mock.AsyncMock(wraps=trigger.create_watcher)) as create_watcher_mock:
        base_order.is_waiting_for_chained_trigger = True
        await base_order.set_as_inactive(trigger)
        # not called for untriggered chained order when they are set as inactive
        create_watcher_mock.assert_not_called()
        base_order.creation_time = 123
        origin_time = base_order.creation_time
        assert base_order.to_dict()[enums.ExchangeConstantsOrderColumns.TIMESTAMP.value] \
            == origin_time
        # base_order.state is None since no state has been associated to it (not initiliazed)
        assert base_order.state is None
        assert base_order.is_initialized is False

        await personal_data.create_as_chained_order(base_order) # will call initialize which will call create_watcher
        assert base_order.creation_time > origin_time   # creation_time got update using current time
        assert base_order.is_waiting_for_chained_trigger is False
        assert base_order.is_created() is True
        assert base_order.is_initialized is True
        assert isinstance(base_order.state, personal_data.OpenOrderState)
        # called for triggered chained order when they are created as triggered chained orders
        create_watcher_mock.assert_awaited_once_with(
            exchange_manager_inst, base_order.symbol, base_order.creation_time
        )

        # use chained order creation time
        assert base_order.to_dict()[enums.ExchangeConstantsOrderColumns.TIMESTAMP.value] \
            == base_order.creation_time


@pytest.mark.asyncio
async def test_create_as_chained_order_bundled_order_no_open_order(trader_simulator):
    config, exchange_manager_inst, trader_inst = trader_simulator

    base_order = personal_data.BuyLimitOrder(trader_inst)
    base_order.update(order_type=enums.TraderOrderType.BUY_LIMIT,
                      symbol="BTC/USDT",
                      current_price=decimal.Decimal("70"),
                      quantity=decimal.Decimal("10"),
                      price=decimal.Decimal("70"))
    base_order.has_been_bundled = True
    # base_order.state is None since no state has been associated to it (not initiliazed)
    assert base_order.state is None
    assert base_order.is_initialized is False

    #no order in order_manager, register as pending order
    assert not exchange_manager_inst.exchange_personal_data.orders_manager.orders
    assert not exchange_manager_inst.exchange_personal_data.orders_manager.pending_creation_orders

    # simulate real trader
    trader_inst.simulate = False
    await personal_data.create_as_chained_order(base_order)
    # did not add it to orders
    assert not exchange_manager_inst.exchange_personal_data.orders_manager.orders
    assert exchange_manager_inst.exchange_personal_data.orders_manager.pending_creation_orders == [base_order]

    assert base_order.is_waiting_for_chained_trigger is False
    # still not initialized
    assert base_order.is_initialized is False


@pytest.mark.asyncio
async def test_create_as_chained_order_bundled_order_no_open_order(trader_simulator):
    config, exchange_manager_inst, trader_inst = trader_simulator

    open_order_1 = personal_data.SellLimitOrder(trader_inst)
    open_order_2 = personal_data.BuyLimitOrder(trader_inst)
    open_order_1.update(order_type=enums.TraderOrderType.SELL_LIMIT,
                        order_id="open_order_1_id",
                        symbol="BTC/USDT",
                        current_price=decimal.Decimal("70"),
                        quantity=decimal.Decimal("10"),
                        price=decimal.Decimal("70"))
    open_order_2.update(order_type=enums.TraderOrderType.BUY_LIMIT,
                        order_id="open_order_2_id",
                        symbol="BTC/USDT",
                        current_price=decimal.Decimal("70"),
                        quantity=decimal.Decimal("10"),
                        price=decimal.Decimal("70"),
                        reduce_only=True)
    await exchange_manager_inst.exchange_personal_data.orders_manager.upsert_order_instance(open_order_1)
    await exchange_manager_inst.exchange_personal_data.orders_manager.upsert_order_instance(open_order_2)
    assert not exchange_manager_inst.exchange_personal_data.orders_manager.pending_creation_orders
    assert exchange_manager_inst.exchange_personal_data.orders_manager.get_all_orders() == [
        open_order_1, open_order_2
    ]

    base_order = personal_data.BuyLimitOrder(trader_inst)
    base_order.update(order_type=enums.TraderOrderType.BUY_LIMIT,
                      order_id="base_order_id",
                      symbol="BTC/USDT",
                      current_price=decimal.Decimal("55"),
                      quantity=decimal.Decimal("10"),
                      price=decimal.Decimal("70"),
                      reduce_only=False)
    base_order.has_been_bundled = True

    # simulate real trader
    trader_inst.simulate = False
    await personal_data.create_as_chained_order(base_order)
    registered_orders = exchange_manager_inst.exchange_personal_data.orders_manager.get_all_orders()
    assert len(registered_orders) == 2
    assert registered_orders[0] is open_order_1
    inserted_pending_order = exchange_manager_inst.exchange_personal_data.orders_manager.get_all_orders()[1]
    assert inserted_pending_order is base_order
    assert inserted_pending_order.reduce_only is True  # inserted_pending_order got preserved and updated
    # did not add it to orders
    assert not exchange_manager_inst.exchange_personal_data.orders_manager.pending_creation_orders

    assert base_order.is_waiting_for_chained_trigger is False
    # still not initialized
    assert base_order.is_initialized is False
    # open_order_2 got cleared
    assert open_order_2.exchange_manager is None


@pytest.mark.asyncio
async def test_create_as_active_order_using_strategy_if_any(trader_simulator):
    config, exchange_manager_inst, trader_inst = trader_simulator
    inactive_order = personal_data.SellLimitOrder(trader_inst)
    group = personal_data.OneCancelsTheOtherOrderGroup(
        "plop", exchange_manager_inst.exchange_personal_data.orders_manager
    )

    with mock.patch.object(order_util, "create_as_active_order_on_exchange", mock.AsyncMock()) as create_as_active_order_on_exchange, \
        mock.patch.object(group.active_order_swap_strategy.__class__, "execute", mock.AsyncMock()) as execute_mock:
        # no strategy
        await personal_data.create_as_active_order_using_strategy_if_any(inactive_order, 123, "callback")
        create_as_active_order_on_exchange.assert_called_once_with(inactive_order, False)
        execute_mock.assert_not_called()
        create_as_active_order_on_exchange.reset_mock()

        # with strategy
        inactive_order.order_group = group
        await personal_data.create_as_active_order_using_strategy_if_any(inactive_order, 123, "callback")
        create_as_active_order_on_exchange.assert_not_called()
        execute_mock.assert_called_once_with(inactive_order, "callback", 123)


@pytest.mark.asyncio
async def test_create_as_active_order_on_exchange_with_sub_calls(trader_simulator):
    config, exchange_manager_inst, trader_inst = trader_simulator
    trader_inst.simulate = False
    trader_inst.enable_inactive_orders = True
    inactive_order = personal_data.SellLimitOrder(trader_inst)
    inactive_order.is_active = False
    exchange_created_order = personal_data.SellLimitOrder(trader_inst)
    exchange_created_order.update(
        order_type=enums.TraderOrderType.SELL_LIMIT,
        order_id="base_order_id",
        symbol="BTC/USDT",
        current_price=decimal.Decimal("55"),
        quantity=decimal.Decimal("10"),
        price=decimal.Decimal("70"),
        reduce_only=False,
        fee={
            enums.FeePropertyColumns.COST.value: 1,
            enums.FeePropertyColumns.CURRENCY.value: "USDT",
        }
    )
    await exchange_manager_inst.exchange_personal_data.orders_manager.upsert_order_instance(inactive_order)
    all_orders = exchange_manager_inst.exchange_personal_data.orders_manager.get_all_orders()
    assert all_orders == [inactive_order]
    with mock.patch.object(exchange_manager_inst.exchange, "create_order", mock.AsyncMock(return_value=exchange_created_order)) as create_order_mock, \
        mock.patch.object(order_factory, "create_order_instance_from_raw", mock.Mock(return_value=exchange_created_order)) as create_order_instance_from_raw_mock:
        active_order = await personal_data.create_as_active_order_on_exchange(inactive_order, True)
        assert active_order is exchange_created_order
        assert active_order.is_active is True
        all_orders = exchange_manager_inst.exchange_personal_data.orders_manager.get_all_orders()
        # inactive order is replaced by active_order
        assert all_orders == [active_order]
        # inactive_order is now active and cleared
        assert inactive_order.is_active is True
        assert inactive_order.exchange_manager is None
        create_order_mock.assert_called_once()
        create_order_instance_from_raw_mock.assert_called_once()


@pytest.mark.asyncio
async def test_update_order_as_inactive_on_exchange_with_sub_calls(trader_simulator):
    config, exchange_manager_inst, trader_inst = trader_simulator
    trader_inst.enable_inactive_orders = True
    trader_inst.simulate = False
    active_order = personal_data.SellLimitOrder(trader_inst)
    active_order.update(
        order_type=enums.TraderOrderType.SELL_LIMIT,
        order_id="base_order_id",
        symbol="BTC/USDT",
        current_price=decimal.Decimal("55"),
        quantity=decimal.Decimal("10"),
        price=decimal.Decimal("70"),
        reduce_only=False,
        active_trigger=personal_data.create_order_price_trigger(active_order, decimal.Decimal("70"), False),
        fee={
            enums.FeePropertyColumns.COST.value: 1,
            enums.FeePropertyColumns.CURRENCY.value: "USDT",
        }
    )
    await exchange_manager_inst.exchange_personal_data.orders_manager.upsert_order_instance(active_order)
    all_orders = exchange_manager_inst.exchange_personal_data.orders_manager.get_all_orders()
    assert all_orders == [active_order]
    assert active_order.active_trigger._trigger_event is None
    with mock.patch.object(exchange_manager_inst.exchange, "cancel_order", mock.AsyncMock(return_value=enums.OrderStatus.CANCELED)) as cancel_order_mock:
        assert await personal_data.update_order_as_inactive_on_exchange(active_order, False) is True
        assert exchange_manager_inst.exchange_personal_data.orders_manager.get_all_orders() == [active_order]
        # order is now inactive
        assert exchange_manager_inst.exchange_personal_data.orders_manager.get_all_orders(active=False) == [active_order]
        assert active_order.is_active is False
        assert active_order.status == enums.OrderStatus.OPEN
        assert active_order.active_trigger is not None
        assert active_order.active_trigger._trigger_event is not None
        cancel_order_mock.assert_called_once()


@pytest.mark.asyncio
async def test_ensure_orders_relevancy_without_positions(trader_simulator):
    config, exchange_manager_inst, trader_inst = trader_simulator
    order_mock = Mock(exchange_manager=exchange_manager_inst, symbol="BTC/USD")
    exchange_manager_inst.exchange_personal_data.positions_manager.positions = {}
    # without positions: doing nothing
    async with personal_data.ensure_orders_relevancy(order=order_mock):
        pass
    # nothing happened


@pytest.mark.asyncio
async def test_ensure_orders_relevancy_with_positions(future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear
    position = personal_data.LinearPosition(trader_inst, default_contract)
    position.symbol = DEFAULT_FUTURE_SYMBOL
    order_mock = Mock(exchange_manager=exchange_manager_inst, symbol=DEFAULT_FUTURE_SYMBOL)
    trader_mock = Mock(cancel_order=AsyncMock())
    group_mock = Mock(on_cancel=AsyncMock())
    to_cancel_order_mock = Mock(trader=trader_mock, order_group=group_mock, status=enums.OrderStatus.OPEN,
                                     is_open=Mock(return_value=True), reduce_only=True,
                                     symbol=DEFAULT_FUTURE_SYMBOL)
    # with positions
    exchange_manager_inst.exchange_personal_data.positions_manager.positions = {"BTC/USDT": position}
    exchange_manager_inst.exchange_personal_data.orders_manager.orders = {"id": to_cancel_order_mock}
    async with personal_data.ensure_orders_relevancy(order=order_mock):
        # no change
        pass
    trader_mock.cancel_order.assert_not_called()
    # with order parameter
    async with personal_data.ensure_orders_relevancy(order=order_mock):
        # changing side
        position.side = "other_side"
    trader_mock.cancel_order.assert_not_called()
    # don't canceled order as position was idle (same situation as open a short from a 0 quantity position,
    # which is considered as long)
    async with personal_data.ensure_orders_relevancy(position=position):
        # changing side
        position.side = "other_other_side"
    trader_mock.cancel_order.assert_not_called()
    # with a non-0 quantity position
    position.quantity = decimal.Decimal("2")
    # with position parameter
    async with personal_data.ensure_orders_relevancy(position=position):
        # changing side
        position.side = "other_side"
    # canceled order
    trader_mock.cancel_order.assert_called_once_with(to_cancel_order_mock)


@pytest.mark.asyncio
async def test_get_order_size_portfolio_percent(trader_simulator):
    config, exchange_manager_inst, trader_inst = trader_simulator
    api.force_set_mark_price(exchange_manager_inst, "BTC/UDST", 1000)
    # no USDT in portfolio
    assert await personal_data.get_order_size_portfolio_percent(
        exchange_manager_inst, decimal.Decimal("0.1"), enums.TradeOrderSide.BUY, "BTC/UDST"
    ) == constants.ZERO
    # 10 BTC in portfolio
    assert await personal_data.get_order_size_portfolio_percent(
        exchange_manager_inst, decimal.Decimal("1"), enums.TradeOrderSide.SELL, "BTC/UDST"
    ) == decimal.Decimal("10")
    assert await personal_data.get_order_size_portfolio_percent(
        exchange_manager_inst, decimal.Decimal("10"), enums.TradeOrderSide.SELL, "BTC/UDST"
    ) == decimal.Decimal("100")
    assert await personal_data.get_order_size_portfolio_percent(
        exchange_manager_inst, decimal.Decimal("11"), enums.TradeOrderSide.SELL, "BTC/UDST"
    ) == decimal.Decimal("100")
    assert await personal_data.get_order_size_portfolio_percent(
        exchange_manager_inst, decimal.Decimal("6.6666"), enums.TradeOrderSide.SELL, "BTC/UDST"
    ) == decimal.Decimal("66.666")

    # 100 USDT in portfolio
    exchange_manager_inst.exchange_personal_data.portfolio_manager.portfolio.portfolio['UDST'].available \
        = decimal.Decimal("100")
    exchange_manager_inst.exchange_personal_data.portfolio_manager.portfolio.portfolio['UDST'].total \
        = decimal.Decimal("100")
    assert await personal_data.get_order_size_portfolio_percent(
        exchange_manager_inst, decimal.Decimal("0.1"), enums.TradeOrderSide.BUY, "BTC/UDST"
    ) == decimal.Decimal("100")
    assert await personal_data.get_order_size_portfolio_percent(
        exchange_manager_inst, decimal.Decimal("0.01"), enums.TradeOrderSide.BUY, "BTC/UDST"
    ) == decimal.Decimal("10")


def test_get_valid_split_orders():
    # example SOL/BTC values from 12/09/2023
    symbol_market = {
        enums.ExchangeConstantsMarketStatusColumns.LIMITS.value: {
            enums.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT.value: {
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT_MIN.value: 0.01,
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT_MAX.value: 90000000.0,
            },
            enums.ExchangeConstantsMarketStatusColumns.LIMITS_COST.value: {
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_COST_MIN.value: 0.0001,
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_COST_MAX.value: 9000000.0
            },
            enums.ExchangeConstantsMarketStatusColumns.LIMITS_PRICE.value: {
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_PRICE_MIN.value: 1e-07,
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_PRICE_MAX.value: 1000.0
            },
        },
        enums.ExchangeConstantsMarketStatusColumns.PRECISION.value: {
            enums.ExchangeConstantsMarketStatusColumns.PRECISION_PRICE.value: 7,
            enums.ExchangeConstantsMarketStatusColumns.PRECISION_AMOUNT.value: 2
        }
    }
    # all valid values
    prices = [
        decimal.Decimal("0.0006858"),
        decimal.Decimal("0.0006908"),
        decimal.Decimal("0.0006938"),
        decimal.Decimal("0.0006958"),
    ]
    assert personal_data.get_valid_split_orders(
        decimal.Decimal("1"),
        prices,
        symbol_market
    ) == ([decimal.Decimal("1") / len(prices)] * len(prices), prices)

    # too small amount for cost: adapt to 3 orders
    assert personal_data.get_valid_split_orders(
        decimal.Decimal("0.5"),
        prices,
        symbol_market
    ) == ([decimal.Decimal("0.5") / 3] * 3, prices[:3])

    # too small amount for cost: adapt to 1 orders
    assert personal_data.get_valid_split_orders(
        decimal.Decimal("0.1"),
        prices,
        symbol_market
    ) == ([decimal.Decimal('0.1')], [decimal.Decimal('0.0006858')])

    # too small amount for cost (because of the lowest price order amount that gets truncated because of
    # exchange precision rules): adapt to 0 orders

    prices = [
        decimal.Decimal("0.0006963"),
        decimal.Decimal("0.0006908"),
        decimal.Decimal("0.0006938"),
        decimal.Decimal("0.0006958"),
    ]
    assert personal_data.get_valid_split_orders(
        decimal.Decimal("0.14985"),
        prices,
        symbol_market,
    ) == ([decimal.Decimal('0.14985')], [decimal.Decimal('0.0006963')])


def test_get_valid_split_orders_with_amount_ratio_per_order():
    # example SOL/BTC values from 12/09/2023
    symbol_market = {
        enums.ExchangeConstantsMarketStatusColumns.LIMITS.value: {
            enums.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT.value: {
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT_MIN.value: 0.01,
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT_MAX.value: 90000000.0,
            },
            enums.ExchangeConstantsMarketStatusColumns.LIMITS_COST.value: {
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_COST_MIN.value: 0.0001,
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_COST_MAX.value: 9000000.0
            },
            enums.ExchangeConstantsMarketStatusColumns.LIMITS_PRICE.value: {
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_PRICE_MIN.value: 1e-07,
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_PRICE_MAX.value: 1000.0
            },
        },
        enums.ExchangeConstantsMarketStatusColumns.PRECISION.value: {
            enums.ExchangeConstantsMarketStatusColumns.PRECISION_PRICE.value: 7,
            enums.ExchangeConstantsMarketStatusColumns.PRECISION_AMOUNT.value: 2
        }
    }

    # all valid values
    input_prices = [
        decimal.Decimal("0.0006858"),
        decimal.Decimal("0.0006908"),
        decimal.Decimal("0.0006938"),
        decimal.Decimal("0.0006958"),
    ]
    amount_ratio_per_order = [
        decimal.Decimal("0.12"),
        decimal.Decimal("0.24"),
        decimal.Decimal("0.36"),
        decimal.Decimal("0.28"),
    ]
    # invalid inputs
    with pytest.raises(ValueError):
        personal_data.get_valid_split_orders(
            decimal.Decimal("2"),
            input_prices,
            symbol_market,
            amount_ratio_per_order=[1, 2, 3]    # not enough values
        )
    with pytest.raises(ValueError):
        personal_data.get_valid_split_orders(
            decimal.Decimal("2"),
            [decimal.Decimal("0.5")],
            symbol_market,
            amount_ratio_per_order=[decimal.Decimal(0)]    # 0 values
        )
    with pytest.raises(ValueError):
        personal_data.get_valid_split_orders(
            decimal.Decimal("2"),
            [decimal.Decimal("0.5")],
            symbol_market,
            amount_ratio_per_order=[decimal.Decimal(-1.2)]    # negative values
        )

    # valid inputs
    volumes, prices = personal_data.get_valid_split_orders(
        decimal.Decimal("2"),
        input_prices,
        symbol_market,
        amount_ratio_per_order=amount_ratio_per_order
    )
    assert (volumes, prices) == (
        [decimal.Decimal("2") * amount_ratio for amount_ratio in amount_ratio_per_order],
        prices
    )
    assert sum(volumes) == decimal.Decimal("2")

    # too small amount for cost: adapt to 2 orders
    amount_ratio_per_order = [decimal.Decimal(33), decimal.Decimal(44), decimal.Decimal(55), decimal.Decimal(66)]
    total_ratio = sum(amount_ratio_per_order[:2])
    volumes, prices = personal_data.get_valid_split_orders(
        decimal.Decimal("0.5"),
        input_prices,
        symbol_market,
        amount_ratio_per_order=amount_ratio_per_order
    )
    assert (volumes, prices) == (
        [decimal.Decimal("0.5") * amount_ratio/total_ratio for amount_ratio in amount_ratio_per_order[:2]],
        prices
    )
    assert sum(volumes) == decimal.Decimal("0.5")

    # too small amount for cost: adapt to 1 order
    volumes, prices = personal_data.get_valid_split_orders(
        decimal.Decimal("0.1"),
        input_prices,
        symbol_market,
        amount_ratio_per_order=amount_ratio_per_order
    )
    assert (volumes, prices) == ([decimal.Decimal('0.1')], [decimal.Decimal('0.0006858')])

    # too small amount for cost (because of the lowest price order amount that gets truncated because of
    # exchange precision rules): adapt to 0 orders

    input_prices = [
        decimal.Decimal("0.0006963"),
        decimal.Decimal("0.0006908"),
        decimal.Decimal("0.0006938"),
        decimal.Decimal("0.0006958"),
    ]
    assert personal_data.get_valid_split_orders(
        decimal.Decimal("0.14985"),
        input_prices,
        symbol_market,
        amount_ratio_per_order=amount_ratio_per_order
    ) == ([decimal.Decimal('0.14985')], [decimal.Decimal('0.0006963')])


def test_get_split_orders_count_and_increment():
    # example SOL/BTC values from 12/09/2023
    symbol_market = {
        enums.ExchangeConstantsMarketStatusColumns.LIMITS.value: {
            enums.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT.value: {
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT_MIN.value: 0.01,
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT_MAX.value: 90000000.0,
            },
            enums.ExchangeConstantsMarketStatusColumns.LIMITS_COST.value: {
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_COST_MIN.value: 0.0001,
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_COST_MAX.value: 9000000.0
            },
            enums.ExchangeConstantsMarketStatusColumns.LIMITS_PRICE.value: {
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_PRICE_MIN.value: 1e-07,
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_PRICE_MAX.value: 1000.0
            },
        },
        enums.ExchangeConstantsMarketStatusColumns.PRECISION.value: {
            enums.ExchangeConstantsMarketStatusColumns.PRECISION_PRICE.value: 7,
            enums.ExchangeConstantsMarketStatusColumns.PRECISION_AMOUNT.value: 2
        }
    }
    # all valid values
    assert personal_data.get_split_orders_count_and_increment(
        decimal.Decimal("0.0006858"),
        decimal.Decimal("0.0006958"),
        decimal.Decimal("1"),
        4,
        symbol_market,
        True
    ) == (4, decimal.Decimal('0.0000025'))

    # too small amount for cost: adapt to 3 orders
    assert personal_data.get_split_orders_count_and_increment(
        decimal.Decimal("0.0006858"),
        decimal.Decimal("0.0006958"),
        decimal.Decimal("0.5"),
        4,
        symbol_market,
        True
    ) == (3, decimal.Decimal('0.000003333333333333333333333333333'))

    # too small amount for cost: adapt to 0 orders
    assert personal_data.get_split_orders_count_and_increment(
        decimal.Decimal("0.0006858"),
        decimal.Decimal("0.0006958"),
        decimal.Decimal("0.1"),
        4,
        symbol_market,
        True
    ) == (0, decimal.Decimal('0'))

    # too small amount for cost (because of the lowest price order amount that gets truncated because of
    # exchange precision rules): adapt to 0 orders
    assert personal_data.get_split_orders_count_and_increment(
        decimal.Decimal("0.0006963"),
        decimal.Decimal("0.0006958"),
        decimal.Decimal("0.14985"),
        4,
        symbol_market,
        False
    ) == (0, decimal.Decimal('0'))


def test_ensure_orders_limit():
    symbol = "BTC/USDT"
    exchange_manager = mock.Mock(
        exchange = mock.Mock(
            get_max_orders_count = mock.Mock(return_value=5),
        ),
        exchange_personal_data = mock.Mock(
            orders_manager = mock.Mock(
                get_open_orders = mock.Mock(
                    return_value=[
                        mock.Mock(order_type=enums.TraderOrderType.BUY_LIMIT),
                        mock.Mock(order_type=enums.TraderOrderType.SELL_LIMIT),
                        mock.Mock(order_type=enums.TraderOrderType.STOP_LOSS),
                        mock.Mock(order_type=enums.TraderOrderType.STOP_LOSS),
                        mock.Mock(order_type=enums.TraderOrderType.STOP_LOSS),
                    ],
                )
            )
        )
    )

    # do not raise
    order_util.ensure_orders_limit(exchange_manager, symbol, [])
    assert exchange_manager.exchange.get_max_orders_count.call_count == 2
    assert exchange_manager.exchange.get_max_orders_count.mock_calls[0].args == (
        symbol, enums.TraderOrderType.SELL_LIMIT
    )
    assert exchange_manager.exchange.get_max_orders_count.mock_calls[1].args == (
        symbol, enums.TraderOrderType.STOP_LOSS
    )
    exchange_manager.exchange_personal_data.orders_manager.get_open_orders.assert_called_once_with(
        symbol=symbol, active=None
    )

    # ok added limit orders
    order_util.ensure_orders_limit(exchange_manager, symbol, [enums.TraderOrderType.BUY_LIMIT]*2)
    order_util.ensure_orders_limit(exchange_manager, symbol, [enums.TraderOrderType.BUY_LIMIT]*3)
    # too many added limit orders
    with pytest.raises(octobot_trading.errors.MaxOpenOrderReachedForSymbolError):
        order_util.ensure_orders_limit(exchange_manager, symbol, [enums.TraderOrderType.BUY_LIMIT]*4)
    with pytest.raises(octobot_trading.errors.MaxOpenOrderReachedForSymbolError):
        order_util.ensure_orders_limit(exchange_manager, symbol, [enums.TraderOrderType.BUY_LIMIT]*10)

    # ok added stop orders
    order_util.ensure_orders_limit(exchange_manager, symbol, [enums.TraderOrderType.STOP_LOSS])
    order_util.ensure_orders_limit(exchange_manager, symbol, [enums.TraderOrderType.STOP_LOSS_LIMIT]*2)
    # too many added stop orders
    with pytest.raises(octobot_trading.errors.MaxOpenOrderReachedForSymbolError):
        order_util.ensure_orders_limit(exchange_manager, symbol, [enums.TraderOrderType.STOP_LOSS]*3)
    with pytest.raises(octobot_trading.errors.MaxOpenOrderReachedForSymbolError):
        order_util.ensure_orders_limit(exchange_manager, symbol, [enums.TraderOrderType.STOP_LOSS_LIMIT]*10)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import pytest
import decimal
import mock

import octobot_trading.personal_data as personal_data


@pytest.fixture
def mock_callback():
    return mock.AsyncMock()


@pytest.fixture
def trigger_price():
    return decimal.Decimal("100.0")


@pytest.fixture
def price_trigger(mock_callback, trigger_price):
    return personal_data.PriceTrigger(
        on_trigger_callback=mock_callback,
        on_trigger_callback_args=("arg1", "arg2"),
        trigger_price=trigger_price,
        trigger_above=True
    )


@pytest.mark.parametrize("current_price,trigger_above,expected", [
    (decimal.Decimal("101.0"), True, True),  # Above trigger price, trigger_above=True
    (decimal.Decimal("99.0"), True, False),  # Below trigger price, trigger_above=True
    (decimal.Decimal("100.0"), True, True),  # Equal trigger price, trigger_above=True
    (decimal.Decimal("101.0"), False, False),  # Above trigger price, trigger_above=False
    (decimal.Decimal("99.0"), False, True),  # Below trigger price, trigger_above=False
    (decimal.Decimal("100.0"), False, True),  # Equal trigger price, trigger_above=False
])
def test_triggers(current_price, trigger_above, expected, mock_callback, trigger_price):
    trigger = personal_data.PriceTrigger(mock_callback, ("arg1", "arg2"), trigger_price, trigger_above)
    assert trigger.triggers(current_price) == expected


def test_init(price_trigger, mock_callback, trigger_price):
    assert price_trigger.trigger_price == trigger_price
    assert price_trigger.trigger_above is True
    assert price_trigger.on_trigger_callback == mock_callback
    assert price_trigger.on_trigger_callback_args == ("arg1", "arg2")
    assert price_trigger._exchange_manager is None
    assert price_trigger._symbol is None


def test_update_from_other_trigger(price_trigger):
    other_trigger = personal_data.PriceTrigger(
        mock.AsyncMock(),
        ("other_arg",),
        decimal.Decimal("200.0"),
        False
    )

    price_trigger.update_from_other_trigger(other_trigger)
    assert price_trigger.trigger_price == decimal.Decimal("200.0")
    assert price_trigger.trigger_above is False
    # Original callback and args should remain unchanged
    assert price_trigger.on_trigger_callback_args == ("arg1", "arg2")

def test_update(price_trigger):
    # Setup
    callback = mock.Mock()

    # Test update without exchange manager (no event creation)
    new_price = decimal.Decimal("150")
    price_trigger.update(trigger_price=new_price)
    assert price_trigger.trigger_price == new_price

    # Setup exchange manager mock
    price_trigger._exchange_manager = mock.Mock()
    price_trigger._symbol = "BTC/USD"
    price_trigger._trigger_event = mock.Mock()

    # Test update with exchange manager and event creation
    newer_price = decimal.Decimal("200")
    min_trigger_time = 1234.56

    with mock.patch.object(price_trigger, '_create_event') as mock_create_event, \
            mock.patch.object(price_trigger, '_clear_event') as mock_clear_event:
        price_trigger.update(
            trigger_price=newer_price,
            min_trigger_time=min_trigger_time,
            update_event=True
        )

        assert price_trigger.trigger_price == newer_price
        mock_clear_event.assert_called_once()
        mock_create_event.assert_called_once_with(min_trigger_time)

    # Test update without event update
    newest_price = decimal.Decimal("250")
    with mock.patch.object(price_trigger, '_create_event') as mock_create_event, \
            mock.patch.object(price_trigger, '_clear_event') as mock_clear_event:
        price_trigger.update(
            trigger_price=newest_price,
            min_trigger_time=min_trigger_time,
            update_event=False
        )

        assert price_trigger.trigger_price == newest_price
        mock_clear_event.assert_not_called()
        mock_create_event.assert_not_called()

    # Test update with same price (should not trigger event updates)
    with mock.patch.object(price_trigger, '_create_event') as mock_create_event, \
            mock.patch.object(price_trigger, '_clear_event') as mock_clear_event:
        price_trigger.update(
            trigger_price=newest_price,
            min_trigger_time=min_trigger_time,
            update_event=True
        )

        assert price_trigger.trigger_price == newest_price
        mock_clear_event.assert_not_called()
        mock_create_event.assert_not_called()


def test_create_event(price_trigger):
    # Setup
    callback = mock.Mock()
    # Mock exchange manager and its components
    exchange_manager = mock.Mock()
    symbol_data = mock.Mock()
    price_events_manager = mock.Mock()

    exchange_manager.exchange_symbols_data.get_exchange_symbol_data.return_value = symbol_data
    symbol_data.price_events_manager = price_events_manager

    # Test event creation
    min_trigger_time = 1234.56
    price_trigger._exchange_manager = exchange_manager
    price_trigger._symbol = "BTC/USD"

    price_trigger._create_event(min_trigger_time)

    # Verify the event was created with correct parameters
    price_events_manager.new_event.assert_called_once_with(
        price_trigger.trigger_price,
        min_trigger_time,
        price_trigger.trigger_above,
        False
    )

    # Verify the event was stored
    assert price_trigger._trigger_event == price_events_manager.new_event.return_value


def test_str_representation(price_trigger):
    expected = (f"PriceTrigger({price_trigger.on_trigger_callback.__name__}): "
                f"trigger_price={price_trigger.trigger_price}, "
                f"trigger_above={price_trigger.trigger_above}")
    assert str(price_trigger) == expected


def test_clear_with_exchange_manager(price_trigger):
    # Setup mock exchange manager and related objects
    mock_price_events_manager = mock.Mock()
    mock_symbol_data = mock.Mock()
    mock_symbol_data.price_events_manager = mock_price_events_manager
    mock_exchange_symbols_data = mock.Mock()
    mock_exchange_symbols_data.get_exchange_symbol_data.return_value = mock_symbol_data
    mock_exchange_manager = mock.Mock()
    mock_exchange_manager.exchange_symbols_data = mock_exchange_symbols_data

    # Setup trigger with mock objects
    price_trigger._exchange_manager = mock_exchange_manager
    price_trigger._symbol = "BTC/USD"
    price_trigger._trigger_event = mock.Mock()

    # Call clear
    price_trigger.clear()

    # Verify that event was removed
    mock_price_events_manager.remove_event.assert_called_once_with(price_trigger._trigger_event)
    assert price_trigger._exchange_manager is None


def test_create_trigger_event(price_trigger):
    # Setup mock objects
    mock_price_events_manager = mock.Mock()
    mock_symbol_data = mock.Mock()
    mock_symbol_data.price_events_manager = mock_price_events_manager
    mock_exchange_symbols_data = mock.Mock()
    mock_exchange_symbols_data.get_exchange_symbol_data.return_value = mock_symbol_data
    mock_exchange_manager = mock.Mock()
    mock_exchange_manager.exchange_symbols_data = mock_exchange_symbols_data

    symbol = "BTC/USD"
    min_trigger_time = 1.0

    # Call create_trigger_event
    price_trigger._create_trigger_event(mock_exchange_manager, symbol, min_trigger_time)

    # Verify the event was created with correct parameters
    mock_price_events_manager.new_event.assert_called_once_with(
        price_trigger.trigger_price,
        min_trigger_time,
        price_trigger.trigger_above,
        False
    )
    assert price_trigger._exchange_manager == mock_exchange_manager
    assert price_trigger._symbol == symbol



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import typing
import decimal

import octobot_commons.symbols as symbol_util

import octobot_trading.constants as constants
import octobot_trading.errors as errors
import octobot_trading.enums as enums
import octobot_trading.personal_data.portfolios.portfolio_value_holder as portfolio_value_holder


class FuturesPortfolioValueHolder(portfolio_value_holder.PortfolioValueHolder):   
    def get_holdings_ratio(
        self, currency, traded_symbols_only=False, include_assets_in_open_orders=False, coins_whitelist=None
    ) -> typing.Optional[decimal.Decimal]:
        positions_manager = self.portfolio_manager.exchange_manager.exchange_personal_data.positions_manager
        total_portfolio_value: decimal.Decimal = self._get_total_holdings_value(
            coins_whitelist=coins_whitelist, traded_symbols_only=traded_symbols_only
        )

        if currency == self.portfolio_manager.reference_market:
            return self._get_holdings_ratio_from_portfolio(
                self.portfolio_manager.reference_market, traded_symbols_only=traded_symbols_only, coins_whitelist=coins_whitelist, include_assets_in_open_orders=include_assets_in_open_orders
            )

        symbol = currency
        if not symbol_util.is_symbol(symbol):
            try:
                symbol = symbol_util.merge_currencies(currency, self.portfolio_manager.reference_market, settlement_asset=self.portfolio_manager.reference_market)
                position = positions_manager.get_symbol_position(symbol, enums.PositionSide.BOTH)
            except errors.ContractExistsError:
                # try to reverse the symbol
                symbol = symbol_util.merge_currencies(
                    self.portfolio_manager.reference_market, currency, settlement_asset=currency
                )
                position = positions_manager.get_symbol_position(symbol, enums.PositionSide.BOTH)
        else:
            position = positions_manager.get_symbol_position(symbol, enums.PositionSide.BOTH)

        if position.is_idle():
            position_value: decimal.Decimal = constants.ZERO
        else:
            position_value: decimal.Decimal = position.margin

        if include_assets_in_open_orders:
            assets_in_open_orders = self._get_total_holdings_in_open_orders(currency)
            current_in_order_value = self.value_converter.evaluate_value(currency, assets_in_open_orders, init_price_fetchers=False)
            position_value += current_in_order_value

        return position_value / total_portfolio_value if total_portfolio_value > constants.ZERO else constants.ZERO


#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import contextlib
import asyncio
import concurrent.futures
import typing
import decimal

import async_channel.enums as channel_enums
import async_channel.consumer

import octobot_commons.channels_name as channels_name
import octobot_commons.constants as common_constants
import octobot_commons.enums as common_enums
import octobot_commons.logging as logging
import octobot_commons.databases as databases
import octobot_commons.configuration as commons_configuration
import octobot_commons.asyncio_tools as asyncio_tools
import octobot_commons.signals as commons_signals

import octobot_trading.enums as enums
import octobot_trading.constants as constants
import octobot_trading.errors as errors
import octobot_trading.util as util
import octobot_trading.exchanges
import octobot_trading.exchanges.exchanges as exchanges
import octobot_trading.exchange_channel as exchanges_channel
import octobot_trading.modes.channel as modes_channel
import octobot_trading.modes.script_keywords as script_keywords
import octobot_trading.modes.mode_activity as mode_activity
import octobot_trading.modes.abstract_trading_mode as abstract_trading_mode
import octobot_trading.modes.modes_util as modes_util
import octobot_trading.storage.util as storage_util


class AbstractTradingModeProducer(modes_channel.ModeChannelProducer):
    TOPIC_TO_CHANNEL_NAME = {
        common_enums.ActivationTopics.FULL_CANDLES.value:
            channels_name.OctoBotTradingChannelsName.OHLCV_CHANNEL.value,
        common_enums.ActivationTopics.IN_CONSTRUCTION_CANDLES.value:
            channels_name.OctoBotTradingChannelsName.KLINE_CHANNEL.value,
        common_enums.ActivationTopics.EVALUATION_CYCLE.value:
            channels_name.OctoBotEvaluatorsChannelsName.MATRIX_CHANNEL.value,
    }
    # declaring timeout at first trigger
    PRODUCER_LOCKS_BY_EXCHANGE_ID = {}  # use to identify exchange-wide actions

    def __init__(self, channel, config, trading_mode, exchange_manager):
        super().__init__(channel)
        # the trading mode instance logger
        self.logger: logging.BotLogger = logging.get_logger(self.__class__.__name__)

        # the trading mode instance
        self.trading_mode: abstract_trading_mode.AbstractTradingMode = trading_mode

        # the global bot config
        self.config: dict[str, typing.Any] = config

        # the trading mode exchange manager
        self.exchange_manager: octobot_trading.exchanges.ExchangeManager = exchange_manager

        # shortcut
        self.exchange_name: str = self.exchange_manager.exchange_name

        # matrix_id shortcut
        self.matrix_id: typing.Optional[str] = None

        # the final eval used by TradingModeConsumers, default value is 0
        self.final_eval: decimal.Decimal = constants.ZERO

        # the producer state used by TradingModeConsumers
        self.state: typing.Optional[enums.EvaluatorStates] = None

        # the consumer instances
        self.evaluator_consumers: list[async_channel.consumer.Consumer] = []
        self.trading_consumers: list[exchanges_channel.ExchangeChannelConsumer] = []

        self.time_frame_filter: typing.Optional[list[common_enums.TimeFrames]] = None

        # Define trading modes default consumer priority level
        self.priority_level: int = channel_enums.ChannelConsumerPriorityLevels.MEDIUM.value

        self.symbol: typing.Optional[str] = None

        self._is_ready_to_trade: typing.Optional[asyncio.Event] = None
        self.on_reload_config()

        # cleared (awaitable) when inside self.trading_mode_trigger
        self._is_trigger_completed: asyncio.Event = asyncio.Event()
        self._is_trigger_completed.set()

        # used to avoid spamming critical issues
        self._sent_critical_notifications: set[str] = set()

        self.last_activity: mode_activity.TradingModeActivity = mode_activity.TradingModeActivity()

    def on_reload_config(self):
        """
        Called at constructor and after the associated trading mode's reload_config.
        Implement if necessary
        """

    def is_cryptocurrency_wildcard(self):
        """
        Should be True only if self.trading_mode.get_is_cryptocurrency_wildcard() is already True
        But can overwritten (with return False) to disable wildcard trigger when get_is_cryptocurrency_wildcard() is True
        :return: True if the mode producer should be triggered by all cryptocurrencies
        """
        return self.trading_mode.get_is_cryptocurrency_wildcard()

    def is_symbol_wildcard(self):
        """
        Should be True only if self.trading_mode.get_is_symbol_wildcard() is already True
        But can overwritten (with return False) to disable wildcard trigger when get_is_symbol_wildcard() is True
        :return: True if the mode producer should be triggered by all symbols
        """
        return self.trading_mode.get_is_symbol_wildcard()

    def is_time_frame_wildcard(self):
        """
        Should be True only if self.trading_mode.get_is_time_frame_wildcard() is already True
        But can overwritten (with return False) to disable wildcard trigger when get_is_time_frame_wildcard() is True
        :return: True if the mode producer should be triggered by all timeframes
        """
        return self.trading_mode.get_is_time_frame_wildcard()

    # noinspection PyArgumentList
    async def start(self) -> None:
        self._is_ready_to_trade = self._is_ready_to_trade or asyncio.Event()
        try:
            await self.inner_start()
        finally:
            self.logger.debug(
                f"Ready to trade on {self.exchange_manager.exchange_name if self.exchange_manager else ''}, "
                f"symbol: {self.trading_mode.symbol if (self.trading_mode and self.trading_mode.symbol) else ''}"
            )
            self._is_ready_to_trade.set()

    def force_is_ready_to_trade(self):
        if self._is_ready_to_trade is None:
            self._is_ready_to_trade = asyncio.Event()
        self._is_ready_to_trade.set()

    def unset_is_ready_to_trade(self):
        if self._is_ready_to_trade is None:
            self._is_ready_to_trade = asyncio.Event()
        if self._is_ready_to_trade.is_set():
            self._is_ready_to_trade.clear()

    async def inner_start(self) -> None:
        """
        Start trading mode channels subscriptions
        """
        registration_topics = self.get_channels_registration()
        if registration_topics:
            trigger_time_frames = self.get_trigger_time_frames()
            currency_filter = self.trading_mode.cryptocurrency \
                if self.trading_mode.cryptocurrency is not None and not self.is_cryptocurrency_wildcard() \
                else common_constants.CONFIG_WILDCARD
            symbol_filter = self.trading_mode.symbol \
                if self.trading_mode.symbol is not None and not self.is_symbol_wildcard() \
                else common_constants.CONFIG_WILDCARD
            self.time_frame_filter = self.trading_mode.time_frame \
                if self.trading_mode.time_frame is not None and self.is_time_frame_wildcard() \
                else [tf.value
                      for tf in self.exchange_manager.exchange_config.get_relevant_time_frames()
                      if tf.value in trigger_time_frames or
                      trigger_time_frames == common_constants.CONFIG_WILDCARD]
            if trigger_time_frames != common_constants.CONFIG_WILDCARD and \
               len(self.time_frame_filter) < len(trigger_time_frames):
                missing_time_frames = [tf for tf in trigger_time_frames if tf not in self.time_frame_filter]
                self.logger.error(f"Missing timeframe to satisfy {trigger_time_frames} required time frames. "
                                  f"Please activate those timeframes {missing_time_frames}")
            self.matrix_id = exchanges.Exchanges.instance().get_exchange(self.exchange_manager.exchange_name,
                                                                         self.exchange_manager.id).matrix_id
            await self._subscribe_to_registration_topic(registration_topics, currency_filter, symbol_filter)
        await self.init_user_inputs(False)
        await self._wait_for_bot_init(
            self._get_config_init_timeout(),
            extra_symbol_topics=self.get_extra_init_symbol_topics(),
            extra_global_topics=self.get_extra_global_topics()
        )

    def get_extra_init_symbol_topics(self) -> typing.Optional[list]:
        # Implement if necessary
        return None

    def get_extra_global_topics(self) -> typing.Optional[list]:
        # Implement if necessary
        return None

    async def _subscribe_to_registration_topic(self, registration_topics, currency_filter, symbol_filter):
        for registration_topic in registration_topics:
            if registration_topic == channels_name.OctoBotEvaluatorsChannelsName.MATRIX_CHANNEL.value:
                # register to matrix channel if necessary
                try:
                    import octobot_evaluators.evaluators.channel as evaluators_channel
                    import octobot_evaluators.enums as evaluators_enums
                    consumer = await evaluators_channel.get_chan(registration_topic, self.matrix_id).new_consumer(
                        callback=self.get_callback(registration_topic),
                        priority_level=self.priority_level,
                        matrix_id=self.matrix_id,
                        cryptocurrency=currency_filter,
                        symbol=symbol_filter,
                        evaluator_type=evaluators_enums.EvaluatorMatrixTypes.STRATEGIES.value,
                        exchange_name=self.exchange_name,
                        # no time_frame filter to allow receiving updates from strategies without timeframes in wildcard
                        time_frame=common_constants.CONFIG_WILDCARD if self.is_time_frame_wildcard()
                        else self.time_frame_filter,
                        supervised=self.exchange_manager.is_backtesting
                    )
                    self.evaluator_consumers.append(
                        (consumer, registration_topic)
                    )
                    self.trading_mode.is_triggered_after_candle_close = False
                except (KeyError, ImportError):
                    self.logger.error(f"Can't connect matrix channel on {self.exchange_name}")
            else:
                # register to trading channels if necessary
                consumer = await exchanges_channel.get_chan(registration_topic, self.exchange_manager.id).new_consumer(
                    callback=self.get_callback(registration_topic),
                    priority_level=self.priority_level,
                    cryptocurrency=currency_filter,
                    symbol=symbol_filter,
                    time_frame=self.time_frame_filter
                )
                self.trading_consumers.append(
                    (consumer, registration_topic)
                )
                if registration_topic == channels_name.OctoBotTradingChannelsName.OHLCV_CHANNEL.value:
                    self.trading_mode.is_triggered_after_candle_close = True

    def get_callback(self, chan_name):
        return {
            channels_name.OctoBotTradingChannelsName.OHLCV_CHANNEL.value: self.ohlcv_callback,
            channels_name.OctoBotTradingChannelsName.KLINE_CHANNEL.value: self.kline_callback,
            channels_name.OctoBotEvaluatorsChannelsName.MATRIX_CHANNEL.value: self.matrix_callback,
        }[chan_name]

    def get_channels_registration(self):
        registration_channels = []
        # Activate on evaluation cycle only by default
        topic = self.trading_mode.trading_config.get(common_constants.CONFIG_ACTIVATION_TOPICS.replace(" ", "_"),
                                                     common_enums.ActivationTopics.EVALUATION_CYCLE.value)
        try:
            registration_channels.append(self.TOPIC_TO_CHANNEL_NAME[topic])
        except KeyError:
            self.logger.error(f"Unknown registration topic: {topic}")
        return registration_channels

    def get_trigger_time_frames(self):
        return self.trading_mode.trading_config.get(common_constants.CONFIG_TRIGGER_TIMEFRAMES,
                                                    common_constants.CONFIG_WILDCARD)

    async def stop(self) -> None:
        """
        Stop trading mode channels subscriptions
        """
        await super().stop()
        if self.exchange_manager is not None:
            for consumer, channel_name in self.evaluator_consumers:
                try:
                    import octobot_evaluators.evaluators.channel as evaluators_channel
                    await evaluators_channel.get_chan(channel_name,
                                                      exchanges.Exchanges.instance().get_exchange(
                                                          self.exchange_manager.exchange_name,
                                                          self.exchange_manager.id).matrix_id
                                                      ).remove_consumer(consumer)
                except (KeyError, ImportError):
                    self.logger.error(f"Can't unregister {channel_name} channel on {self.exchange_name}")
            for consumer, channel_name in self.trading_consumers:
                try:
                    await exchanges_channel.get_chan(channel_name, self.exchange_manager.id).remove_consumer(consumer)
                except (KeyError, ImportError):
                    self.logger.error(f"Can't unregister {channel_name} channel on {self.exchange_name}")
            self.delete_producer_exchange_wide_lock(self.exchange_manager)
        self.flush()

    def flush(self) -> None:
        """
        Flush all instance objects reference
        """
        self.trading_mode = None # type: ignore
        self.exchange_manager = None # type: ignore
        self.evaluator_consumers = []
        self.trading_consumers = []

    async def ohlcv_callback(self, exchange: str, exchange_id: str, cryptocurrency: str, symbol: str,
                             time_frame: str, candle: dict):
        self.logger.error(f"ohlcv_callback is registered but not implemented")

    async def kline_callback(self, exchange: str, exchange_id: str, cryptocurrency: str, symbol: str,
                             time_frame, kline: dict):
        self.logger.error(f"kline_callback is registered but not implemented")

    async def matrix_callback(self, matrix_id, evaluator_name, evaluator_type,
                              eval_note, eval_note_type, exchange_name, cryptocurrency, symbol, time_frame) -> None:
        """
        Called when a strategy updates the matrix
        :param matrix_id: the matrix_id
        :param evaluator_name: the evaluator name, should be the strategy name
        :param evaluator_type: the evaluator type, should be EvaluatorMatrixTypes.STRATEGIES.value
        :param eval_note: the eval note, should be the strategy eval note
        :param eval_note_type: the eval note type
        :param exchange_name: the exchange name
        :param cryptocurrency: the cryptocurrency
        :param symbol: the symbol
        :param time_frame: the time frame
        """
        if time_frame is None or time_frame in self.time_frame_filter:
            await self.finalize(exchange_name=exchange_name, matrix_id=matrix_id, cryptocurrency=cryptocurrency,
                                symbol=symbol, time_frame=time_frame,
                                trigger_source=common_enums.TriggerSource.EVALUATION_MATRIX.value)

    @modes_util.enabled_trader_only()
    async def finalize(self, exchange_name: str,
                       matrix_id: str,
                       cryptocurrency: str = None,
                       symbol: str = None,
                       time_frame=None,
                       trigger_source: str = common_enums.TriggerSource.UNDEFINED.value
                       ) -> None:
        """
        Finalize evaluation
        """
        if exchange_name != self.exchange_name:
            # Do nothing if not its exchange
            return
        await self.trigger(matrix_id, cryptocurrency, symbol, time_frame, trigger_source)

    async def trigger(self, matrix_id: str = None, cryptocurrency: str = None, symbol: str = None, time_frame=None,
                      trigger_source: str = common_enums.TriggerSource.UNDEFINED.value) -> None:
        """
        Called by finalize and MANUAL_TRIGGER user command. Override if necessary
        """
        try:
            async with self.trading_mode_trigger(), self.trading_mode.remote_signal_publisher(symbol):
                await self.set_final_eval(matrix_id=matrix_id,
                                          cryptocurrency=cryptocurrency,
                                          symbol=symbol,
                                          time_frame=time_frame,
                                          trigger_source=trigger_source)
        except errors.InitializingError as e:
            self.logger.exception(
                e,
                True,
                f"Ignored signal: exchange: {self.exchange_manager.exchange_name} symbol: {symbol}, "
                f"time_frame: {time_frame}. "
                f"Trading mode is not yet ready to trade, OctoBot is still initializing and fetching required data."
            )

    async def wait_for_trigger_completion(self, timeout):
        if self._is_trigger_completed.is_set():
            return
        await asyncio.wait_for(self._is_trigger_completed.wait(), timeout=timeout)

    @contextlib.asynccontextmanager
    async def trading_mode_trigger(self, skip_health_check=False):
        try:
            self._is_trigger_completed.clear()
            if not self._is_ready_to_trade.is_set():
                if self.exchange_manager.is_backtesting:
                    raise asyncio.TimeoutError(f"Trading mode producer has to be started in backtesting")
                self.logger.debug("Waiting for orders initialization to proceed")
                try:
                    await asyncio.wait_for(self._is_ready_to_trade.wait(), self._get_config_init_timeout())
                except asyncio.TimeoutError as e:
                    raise errors.InitializingError() from e
                self.logger.debug("Order initialized")
            if self.trading_mode.is_health_check_required() and not skip_health_check:
                await self.trading_mode.health_check([], {})
            yield
        except errors.InitializingError:
            raise
        except errors.UnreachableExchange as e:
            self.logger.warning(f"Error when calling trading mode: {e}")
        except AttributeError:
            if self._is_ready_to_trade is None:
                raise AttributeError(f"{self.__class__.__name__} has to be started. self._is_ready_to_trade is None")
            raise
        except Exception as e:
            self.logger.exception(e, True, f"Error when calling trading mode: {e}")
        finally:
            self._is_trigger_completed.set()
            await self.post_trigger()

    async def post_trigger(self):
        pass

    async def set_final_eval(self, matrix_id: str, cryptocurrency: str, symbol: str, time_frame,
                             trigger_source: str) -> None:
        """
        Called to calculate the final note or state => when any notification appears
        """
        raise NotImplementedError("set_final_eval not implemented")

    async def submit_trading_evaluation(
        self, cryptocurrency, symbol, time_frame,
        final_note=constants.ZERO,
        state=enums.EvaluatorStates.NEUTRAL,
        data=None,
        dependencies: typing.Optional[commons_signals.SignalDependencies] = None
    ) -> None:
        await self.send(trading_mode_name=self.trading_mode.get_name(),
                        cryptocurrency=cryptocurrency,
                        symbol=symbol,
                        time_frame=time_frame,
                        final_note=final_note,
                        state=state.value,
                        data=data if data is not None else {},
                        dependencies=dependencies)

    @classmethod
    def get_should_cancel_loaded_orders(cls) -> bool:
        """
        Called by cancel_symbol_open_orders => return true if OctoBot should cancel all orders for a symbol including
        orders already existing when OctoBot started up
        """
        raise NotImplementedError("get_should_cancel_loaded_orders not implemented")

    def _get_to_cancel_orders(
        self, symbol=None, side=None, tag=None, active=None, exchange_order_ids=None
    ) -> list:
        cancel_loaded_orders = self.get_should_cancel_loaded_orders()
        return [
            order
            for order in self.exchange_manager.exchange_personal_data.orders_manager.get_open_orders(
                symbol=symbol, tag=tag, active=active
            )
            if (
                not (order.is_cancelled() or order.is_closed())
                and (cancel_loaded_orders or order.is_from_this_octobot)
                and (side is None or (side is order.side))
                and (exchange_order_ids is None or (order.exchange_order_id in exchange_order_ids))
            )
        ]

    async def _cancel_orders(
        self, orders: list, 
        dependencies: typing.Optional[commons_signals.SignalDependencies] = None
    ) -> tuple[bool, typing.Optional[commons_signals.SignalDependencies]]:
        cancelled = False
        failed_to_cancel = False
        cancelled_dependencies = commons_signals.SignalDependencies()
        for order in orders:
            try:
                is_cancelled, new_dependencies = await self.trading_mode.cancel_order(
                    order, dependencies=dependencies
                )
                if is_cancelled:
                    cancelled = True
                    cancelled_dependencies.extend(new_dependencies)
                else:
                    failed_to_cancel = True
            except errors.OctoBotExchangeError as err:
                # do not propagate exchange error when canceling order
                self.logger.exception(err, True, f"Error when cancelling order [{order}]: {err}")
                failed_to_cancel = True
        return (cancelled and not failed_to_cancel), cancelled_dependencies or None

    @modes_util.enabled_trader_only(disabled_return_value=(False, None))
    async def cancel_symbol_open_orders(
        self, symbol, side=None, tag=None, active=None, exchange_order_ids=None,
        dependencies: typing.Optional[commons_signals.SignalDependencies] = None
    ) -> tuple[bool, typing.Optional[commons_signals.SignalDependencies]]:
        """
        Cancel all symbol open orders
        """
        return await self._cancel_orders(
            self._get_to_cancel_orders(
                symbol=symbol, side=side, tag=tag, active=active, exchange_order_ids=exchange_order_ids
            ), dependencies
        )

    @modes_util.enabled_trader_only(disabled_return_value=(False, None))
    async def apply_cancel_policies(
        self, symbol=None, side=None, tag=None, exchange_order_ids=None, active=None,
        dependencies: typing.Optional[commons_signals.SignalDependencies] = None
    ) -> tuple[bool, typing.Optional[commons_signals.SignalDependencies]]:
        """
        Cancel all orders that should be according to their cancel policies
        """
        if to_cancel_orders := self.exchange_manager.exchange_personal_data.orders_manager.get_orders_to_cancel_from_policies(
            self._get_to_cancel_orders(
                symbol=symbol, side=side, tag=tag, active=active, exchange_order_ids=exchange_order_ids
            )
        ):
            self.logger.info(f"Cancelling {len(to_cancel_orders)} orders from cancel policies")
            return await self._cancel_orders(to_cancel_orders, dependencies)
        return False, None

    def all_databases(self):
        provider = databases.RunDatabasesProvider.instance()
        account_type = storage_util.get_account_type_suffix_from_exchange_manager(self.exchange_manager)
        return {
            common_enums.RunDatabases.RUN_DATA_DB.value: provider.get_run_db(self.trading_mode.bot_id),
            common_enums.RunDatabases.ORDERS_DB.value:
                provider.get_orders_db(self.trading_mode.bot_id, account_type, self.exchange_name),
            common_enums.RunDatabases.TRADES_DB.value:
                provider.get_trades_db(self.trading_mode.bot_id, account_type, self.exchange_name),
            common_enums.RunDatabases.TRANSACTIONS_DB.value:
                provider.get_transactions_db(self.trading_mode.bot_id, account_type, self.exchange_name),
            self.trading_mode.symbol:
                provider.get_symbol_db(self.trading_mode.bot_id, self.exchange_name, self.trading_mode.symbol)
                if self.trading_mode.symbol else None,
        }

    async def _wait_for_symbol_init(self, symbol, time_frame, timeout) -> bool:
        try:
            await util.wait_for_topic_init(self.exchange_manager, timeout,
                                           common_enums.InitializationEventExchangeTopics.CANDLES.value, symbol)
            if self.exchange_manager.is_future:
                await util.wait_for_topic_init(self.exchange_manager, timeout,
                                               common_enums.InitializationEventExchangeTopics.CONTRACTS.value, symbol)
            return True
        except (asyncio.TimeoutError, concurrent.futures.TimeoutError):
            self.logger.error(f"Initialization took more than {timeout} seconds")
        return False

    async def _wait_for_bot_init(self, timeout, extra_symbol_topics: list=None, extra_global_topics: list=None) -> bool:
        try:
            init_symbols = self.trading_mode.get_init_symbols()
            symbol_topics = [common_enums.InitializationEventExchangeTopics.ORDERS.value] + (extra_symbol_topics or [])
            if self.trading_mode.REQUIRE_TRADES_HISTORY:
                symbol_topics.append(common_enums.InitializationEventExchangeTopics.TRADES.value)
            for symbol_topic in symbol_topics:
                self.logger.debug(f"Trading mode [{self.exchange_manager.exchange_name}] start complete. "
                                  f"1/2: Now waiting for {symbol_topic} (for {init_symbols}) full initialisation.")
                await util.wait_for_topic_init(
                    self.exchange_manager, timeout, symbol_topic, symbols=init_symbols
                )
            global_topics = [common_enums.InitializationEventExchangeTopics.BALANCE.value] + (extra_global_topics or [])
            for global_topic in global_topics:
                self.logger.debug(f"Trading mode [{self.exchange_manager.exchange_name}] start complete. "
                                  f"2/2: Now waiting for {global_topics} full initialisation.")
                await util.wait_for_topic_init(self.exchange_manager, timeout, global_topic)
            self.logger.debug(
                f"Trading mode requirements init complete: {', '.join(t for t in symbol_topics)} initialisation completed."
            )
            return True
        except (asyncio.TimeoutError, concurrent.futures.TimeoutError):
            self.logger.error(f"Initialization took more than {timeout} seconds")
        return False

    async def init_user_inputs(self, should_clear_inputs):
        if should_clear_inputs:
            await commons_configuration.clear_user_inputs(
                databases.RunDatabasesProvider.instance().get_run_db(self.trading_mode.bot_id)
            )
        await self._register_and_apply_required_user_inputs(
            script_keywords.get_base_context(self.trading_mode, init_call=True)
        )

    async def _register_and_apply_required_user_inputs(self, context):
        if self.trading_mode.ALLOW_CUSTOM_TRIGGER_SOURCE:
            # register activating topics user input
            activation_topic_values = [
                common_enums.ActivationTopics.EVALUATION_CYCLE.value,
                common_enums.ActivationTopics.FULL_CANDLES.value,
                common_enums.ActivationTopics.IN_CONSTRUCTION_CANDLES.value
            ]
            await script_keywords.get_activation_topics(
                context,
                common_enums.ActivationTopics.EVALUATION_CYCLE.value,
                activation_topic_values
            )
        if self.trading_mode.is_updating_exchange_settings(context):
            try:
                await self._apply_exchange_side_config(context)
            except Exception as err:
                # TODO important error to display
                self.logger.exception(err, True, f"Error when applying exchange side config: {err}")

    async def _apply_exchange_side_config(self, context):
        # can be slow, call in a task if necessary
        if context.exchange_manager.is_future:
            if not self._is_ready_to_trade.is_set():
                await util.wait_for_topic_init(
                    self.exchange_manager, self._get_config_init_timeout(),
                    common_enums.InitializationEventExchangeTopics.CONTRACTS.value,
                    symbols=self.trading_mode.get_init_symbols()
                )
            await script_keywords.set_leverage(context, await script_keywords.user_select_leverage(context))

    async def _wait_for_symbol_prices_and_profitability_init(self, timeout) -> bool:
        try:
            await util.wait_for_topic_init(
                self.exchange_manager, timeout, common_enums.InitializationEventExchangeTopics.PRICE.value,
                symbols=self.trading_mode.get_init_symbols()
            )
            await util.wait_for_topic_init(self.exchange_manager, timeout,
                                           common_enums.InitializationEventExchangeTopics.PROFITABILITY.value)
        except (asyncio.TimeoutError, concurrent.futures.TimeoutError):
            self.logger.error(f"Symbol price initialization took more than {timeout} seconds")
        return False

    def _get_config_init_timeout(self) -> int:
        if self.exchange_manager and self.exchange_manager.is_backtesting:
            return 1    # in backesting, init should be almost instant. If it's not, wait for yp to 1 second
        return 1 * common_constants.MINUTE_TO_SECONDS  # let time for orders to be fetched before

    @classmethod
    def producer_exchange_wide_lock(cls, exchange_manager) -> asyncio_tools.RLock():
        try:
            return cls.PRODUCER_LOCKS_BY_EXCHANGE_ID[exchange_manager.id]
        except KeyError:
            lock = asyncio_tools.RLock()
            cls.PRODUCER_LOCKS_BY_EXCHANGE_ID[exchange_manager.id] = lock
            return lock

    @classmethod
    def delete_producer_exchange_wide_lock(cls, exchange_manager):
        if exchange_manager.id in cls.PRODUCER_LOCKS_BY_EXCHANGE_ID:
            cls.PRODUCER_LOCKS_BY_EXCHANGE_ID.pop(
                exchange_manager.id, None
            )

    async def sent_once_critical_notification(self, title: str, description: str):
        if description not in self._sent_critical_notifications:
            try:
                import octobot_services.api as services_api
                import octobot_services.enums as services_enum
                self.logger.error(f"Critical error: {title}: {description}")
                await services_api.send_notification(services_api.create_notification(
                    description, title=title, markdown_text=description,
                    level=services_enum.NotificationLevel.CRITICAL,
                    category=services_enum.NotificationCategory.GLOBAL_INFO
                ))
            except ImportError as e:
                self.logger.exception(e, True, f"Impossible to send notification: {e}")
            finally:
                self._sent_critical_notifications.add(description)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import mock
import pytest

import octobot_trading.personal_data.orders.cancel_policies.order_cancel_policy as order_cancel_policy_import


class TestOrderCancelPolicy:
    """Test the base OrderCancelPolicy class"""

    def test_should_cancel_not_implemented(self):
        """Test that should_cancel raises NotImplementedError in base class"""
        policy = order_cancel_policy_import.OrderCancelPolicy()
        order_mock = mock.Mock()
        
        with pytest.raises(NotImplementedError):
            policy.should_cancel(order_mock)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.constants
import octobot_trading.exchanges as exchanges


def is_channel_managed_by_websocket(exchange_manager, channel):
    return (
        not exchange_manager.rest_only
        and exchange_manager.has_websocket
        and not exchange_manager.is_backtesting
        and channel in octobot_trading.constants.WEBSOCKET_FEEDS_TO_TRADING_CHANNELS
        and exchange_manager.exchange_web_socket is not None
        and any([
            exchange_manager.exchange_web_socket.is_feed_available(feed)
            for feed in octobot_trading.constants.WEBSOCKET_FEEDS_TO_TRADING_CHANNELS[channel]
        ])
    )


def is_channel_fully_managed_by_websocket(exchange_manager, channel):
    return (
        (
            channel not in octobot_trading.constants.ALWAYS_STARTED_REST_PRODUCER_CHANNELS
        ) and (
            not any([
                # no associated feed is related to a time frame
                exchange_manager.exchange_web_socket.is_time_frame_related_feed(feed)
                for feed in octobot_trading.constants.WEBSOCKET_FEEDS_TO_TRADING_CHANNELS[channel]
            ])
            or all([
                # all required time frames are supported
                exchange_manager.exchange_web_socket.is_time_frame_supported(time_frame)
                for time_frame in exchange_manager.exchange_config.available_time_frames
            ])
        ) and (
            not exchange_manager.exchange_config.get_forced_updater_channel_specs(channel)
        )
    )


def is_websocket_feed_requiring_init(exchange_manager, channel):
    return any([exchange_manager.exchange_web_socket.is_feed_requiring_init(feed)
                for feed in octobot_trading.constants.WEBSOCKET_FEEDS_TO_TRADING_CHANNELS[channel]])


async def search_and_create_websocket(exchange_manager):
    ws_exchange_class = exchanges.search_websocket_class(exchanges.WebSocketExchange, exchange_manager)
    if ws_exchange_class is not None:
        await _create_websocket(exchange_manager, exchanges.WebSocketExchange.__name__, ws_exchange_class)


async def _create_websocket(exchange_manager, websocket_class_name, ws_exchange_class):
    try:
        exchange_manager.exchange_web_socket = ws_exchange_class(exchange_manager.config, exchange_manager)
        await _init_websocket(exchange_manager)
        if exchange_manager.exchange_web_socket.is_websocket_running:
            exchange_manager.logger.info(
                f"{ws_exchange_class.get_name()} connecting to {exchange_manager.exchange.name.capitalize()}"
            )
        else:
            exchange_manager.has_websocket = False
    except Exception as e:
        exchange_manager.logger.error(f"Fail to init websocket for {websocket_class_name} "
                                      f"({exchange_manager.exchange.name}): {e}")
        exchange_manager.exchange_web_socket = None
        exchange_manager.has_websocket = False
        raise e


async def _init_websocket(exchange_manager):
    await exchange_manager.exchange_web_socket.init_websocket(exchange_manager.exchange_config.available_time_frames,
                                                              exchange_manager.exchange_config.traded_symbol_pairs,
                                                              exchange_manager.tentacles_setup_config)

    await exchange_manager.exchange_web_socket.start_sockets()

    exchange_manager.has_websocket = exchange_manager.exchange_web_socket.is_websocket_running
